/*
 *  fitsCmds.c --
 *
 *     This holds the handlers for all of the fitsObj commands
 *
 */

#include "fitsTclInt.h"


/*
 * ------------------------------------------------------------
 *
 * fitsDispatch --
 *
 *    This is the dispatch routine for the Fits objects
 *
 *   Results:
 *      Depends on argv[1].
 *
 *   Side Effects:
 *      Depends on argv[1].
 *
 * ------------------------------------------------------------
 *
 */
int fitsDispatch( ClientData clientData,
		  Tcl_Interp *interp,
		  int argc,
		  char *argv[] )
{
  
   static char *commandList =
      "Available commands:\n"
      "close  - close the file and delete this object\n"
      "move ?+/-?n  - move to HDU #n or forward/backward +/-n HDUs\n"
      "dump ?-s/-e/-l?  - return contents of the CHDU's header in various formats\n"
      "info  - get information about the CHDU \n"
      "get   - get various data from CHDU\n"
      "put   - change contents of CHDU: keywords or extension data\n"
      "insert- insert KEYWORDs, COLUMNs, ROWs, or HDUs \n"
      "delete- delete KEYWORDs, COLUMNs, ROWs, or HDUs \n"
      "load  - load image and table data into variables or pointers \n"
      "free  - free loaded data. **If the address is not the right one\n"
      "          returned from \"load xxx\", a core dump will occur** \n"
      "flush ?clear?  - flush dirty buffers to disk (also clear buffer contents?) \n"
      "copy filename - copy the CHDU to a new file\n"
      "sascii- save extension contents to an ascii file \n"
      "sort  - sort the CHDU according to supplied parameters \n"
      "add   - Append new columns and rows to table.  Column may be filled\n"
      "        with the results of a supplied arithmetic expression\n"
      "append filename - Append current HDU to indicated fits file\n"
      "create 1dhisto|2dhisto ...\n"
      "      - Create 1D or 2D histogram from table columns with or without\n"
      "        a weighting column.  Pixel scalings (eg, WCS) are converted.\n"
      "checksum update|verify - Update or verify checksum keywords of the\n"
      "                         current HDU.  Verify: 1=good, -1=bad, 0=none\n"
      ;
   
   int i;
   FitsFD *curFile = (FitsFD *) clientData;
   struct {
      char *cmd;
      int (*fct)(FitsFD*,int,char*[]);
   } cmdLookup[] = {
      { "close",  fitsTcl_close  },
      { "move",   fitsTcl_move   },
      { "dump",   fitsTcl_dump   },
      { "info",   fitsTcl_info   },
      { "get",    fitsTcl_get    },
      { "put",    fitsTcl_put    },
      { "insert", fitsTcl_insert },
      { "delete", fitsTcl_delete },
      { "load",   fitsTcl_load   },
      { "free",   fitsTcl_free   },
      { "flush",  fitsTcl_flush  },
      { "copy",   fitsTcl_copy   },
      { "sascii", fitsTcl_sascii },
      { "sort",   fitsTcl_sort   },
      { "add",    fitsTcl_add    },
      { "append", fitsTcl_append },
      { "create", fitsTcl_create },
      { "checksum", fitsTcl_checksum },
      { "", NULL }
   };
   
   /*
    *  If there are no arguments, return the help string
    */
   
   if( argc==1 ) {
      Tcl_SetResult(interp,commandList,TCL_STATIC);
      return TCL_OK;
   }
   
   /*
    *  Search for the command and call its handler
    */
   
   for( i=0; cmdLookup[i].cmd[0]; i++ ) {
      if( !strcmp( cmdLookup[i].cmd, argv[1] ) )
	 return (*cmdLookup[i].fct)(curFile, argc, argv);
   }
   
   /*
    *  NO SUCH COMMAND...  Error
    */
   
   Tcl_SetResult(interp, "Unrecognized command\n", TCL_STATIC);
   Tcl_AppendResult(interp, commandList);
   return TCL_ERROR;
}



/**********************
 *
 *   Command Handlers....
 *
 **********************/


/******************************************************************
 *                             Close
 ******************************************************************/

int fitsTcl_close( FitsFD *curFile, int argc, char *argv[] )
{
   char varHandle[255];

   if ( argc != 2 ) {
      Tcl_SetResult(curFile->interp,
		    "Wrong number of args: expected fits close",TCL_STATIC);
      return TCL_ERROR;
   }
   strcpy(varHandle, curFile->handleName);
   strcat(varHandle, FITS_HANDLE_END );
   return Tcl_UnsetVar(curFile->interp, varHandle, TCL_GLOBAL_ONLY);
}


/******************************************************************
 *                             Move
 ******************************************************************/

int fitsTcl_move( FitsFD *curFile, int argc, char *argv[] )
{
   static char *moveList = "\n"
      "move nmove - moves the CHDU: \n"
      "             nmove = +- -> relative move, otherwise absolute\n"
      "             returns hdutype\n";

   char result[10];
   int nmove;
   int mSilent=0;
   int status=0;

   if ( 3 > argc ) {
      Tcl_SetResult(curFile->interp, moveList, TCL_STATIC);
      return TCL_OK;
   } 

   /* Convert the nmove argument */

   if( Tcl_GetInt(curFile->interp,argv[2],&nmove) != TCL_OK ) {
      Tcl_SetResult(curFile->interp,"Wrong type for nmove",TCL_STATIC);
      return TCL_ERROR;
   }

   if( argc == 4 ) {
      if( !strcmp(argv[3], "-s") ) {
	 mSilent = 1;
      } else {
	 Tcl_SetResult(curFile->interp, "fitsTcl Error: "
		 "unkown option: -s for load without read header", TCL_STATIC);
	 return TCL_ERROR;
      }
   }
  
   if( mSilent ) {
      
      if ( strchr(argv[2],'+') ) {
	 status = fitsJustMoveHDU(curFile, nmove, 1);
      } else if ( strchr(argv[2],'-') ) {
	 status = fitsJustMoveHDU(curFile, nmove,-1);
      } else {
	 status = fitsJustMoveHDU(curFile, nmove, 0);
      }

   } else {
      
      if ( strchr(argv[2],'+') ) {
	 status = fitsMoveHDU(curFile, nmove, 1);
      } else if ( strchr(argv[2],'-') ) {
	 status = fitsMoveHDU(curFile, nmove,-1);
      } else {
	 status = fitsMoveHDU(curFile, nmove, 0);
      }

   }
   
   if ( status ) {
      return TCL_ERROR;
   }

   /* Put the hdutype in the result */
   sprintf(result,"%d",curFile->hduType);
   Tcl_SetResult(curFile->interp, result, TCL_VOLATILE);
   return TCL_OK;
}


/******************************************************************
 *                             Dump
 ******************************************************************/

int fitsTcl_dump( FitsFD *curFile, int argc, char *argv[] )
{
   int status;

   if( argc == 2 ) {

      status = fitsDumpHeader(curFile);

   } else {

      if( !strcmp("-l",argv[2]) ) {
	 status = fitsDumpKwdsToList(curFile);
      } else if( !strcmp("-s",argv[2]) ) {
	 status = fitsDumpHeaderToKV(curFile);
      } else if( !strcmp("-e",argv[2]) ) {
	 status = fitsDumpHeaderToCard(curFile);
      } else {
	 Tcl_SetResult(curFile->interp,
		       "Usage: fitsFile dump ?-s/-e/-l?", TCL_STATIC);
	 return TCL_ERROR;
      }

   }

   return status;
}


/******************************************************************
 *                             Info
 ******************************************************************/

int fitsTcl_info( FitsFD *curFile, int argc, char *argv[] )
{
   static char *infoList = "\n"
      "Available Commands:\n"
      "\n"
      "info chdu    - returns the CHDU\n"
      "info nhdu    - returns the total number of hdu in the file\n"
      "info filesize- returns the size of the file(in unit of 2880 byte)\n"
      "info hdutype - returns the type of the CHDU\n"
      "info imgdim  - returns the image dimension of the CHDU \n"
      "info ncols   - returns the number of columns in the CHDU\n"
      "info nrows   - returns the number of rows in the CHDU\n"
      "info nkwds   - returns the number of keywords in the CHDU\n"
      "info column ?-exact? ?colNames? \n"
      "               with no argument, lists the columns,\n"
      "               otherwise gives more info about columns in colName\n"
      "            ?-minmax? colName firstElement ?rowRange? \n"
      "               min and max\n"
      "            ?-stat? colName firstElement ?rowRange? \n"
      "               statistics about the indicated column\n"
      "\n";
   
   int i, j, felem, numRange, range[FITS_MAXRANGE][2];
   int numCols, colTypes[FITS_COLMAX], colNums[FITS_COLMAX], strSize[FITS_COLMAX];
   int status = 0;
   char result[32];
   char tmpStr[3][FLEN_VALUE];  /*  Some general purpose string buffers  */
   char *mrgList[9], *pattern, *tmpStrPtr;
   char errMsg[256], **colList;
   Tcl_DString concatList;

   if( argc < 3 ) {
      Tcl_SetResult(curFile->interp, infoList, TCL_STATIC);
      return TCL_OK;
   }

   /* check if the chdu has been loaded or not */

   if( curFile->CHDUInfo.table.loadStatus != 1 ) {

      Tcl_SetResult(curFile->interp,
		    "You need to load the CHDU first", TCL_STATIC);
      return TCL_ERROR;

   }

   if( !strcmp("chdu",argv[2] ) ) {

      sprintf(result,"%d",curFile->chdu);
      Tcl_SetResult(curFile->interp, result, TCL_VOLATILE);

   } else if( !strcmp("filesize",argv[2]) ) {

      sprintf(result,"%ld",curFile->fptr->Fptr->filesize/2880);
      Tcl_SetResult(curFile->interp, result, TCL_VOLATILE);

   } else if( !strcmp("hdutype",argv[2]) ) {

      switch ( curFile->hduType ) {
      case IMAGE_HDU:
	 if( curFile->chdu )
	    tmpStrPtr = "Image extension";
	 else
	    tmpStrPtr = "Primary array";
	 break;
      case ASCII_TBL:
	 tmpStrPtr = "ASCII Table";
	 break;
      case BINARY_TBL:
	 tmpStrPtr = "Binary Table";
	 break;
      default:
	 Tcl_SetResult(curFile->interp, "Unsupported hdu type", TCL_STATIC);
	 return TCL_ERROR;
      }

      Tcl_SetResult(curFile->interp, tmpStrPtr, TCL_STATIC);

   } else if( !strcmp("nhdu", argv[2]) ) {
      int nhdu;

      ffthdu(curFile->fptr, &nhdu, &status);
      if( status ) {
	 dumpFitsErrStack(curFile->interp, status);
	 return TCL_ERROR;
      }
      sprintf(result, "%d", nhdu);
      Tcl_SetResult( curFile->interp, result, TCL_VOLATILE );

   } else if( !strcmp("nkwds",argv[2] ) ) {

      sprintf(result, "%-d", curFile->numKwds);
      Tcl_SetResult( curFile->interp, result, TCL_VOLATILE );

   } else if( !strcmp("ncols",argv[2] ) ) {

      if (curFile->hduType == IMAGE_HDU ) {
	 Tcl_SetResult( curFile->interp,
			"No columns for an Image extension", TCL_STATIC);
	 return TCL_ERROR;
      }
      sprintf(result, "%d", curFile->CHDUInfo.table.numCols);
      Tcl_SetResult( curFile->interp, result, TCL_VOLATILE );
      
   } else if( !strcmp("nrows",argv[2] ) ) {
      
      if (curFile->hduType == IMAGE_HDU ) {
	 Tcl_SetResult( curFile->interp,
			"No rows for an Image extension", TCL_STATIC );
	 return TCL_ERROR;
      }
      sprintf(result,"%ld",curFile->CHDUInfo.table.numRows);
      Tcl_SetResult( curFile->interp, result, TCL_VOLATILE );
      
   } else if( !strcmp("column",argv[2] ) ) { 
	 
      if( curFile->hduType == IMAGE_HDU ) {
	 Tcl_SetResult( curFile->interp,
			"No Columns in an image extension", TCL_STATIC );
	 return TCL_ERROR;
      }
      
      if( argc == 3 ) {

	 /***********************************
	  *  Return a list of column names  *
	  ***********************************/

	 for ( i = 0; i < curFile->CHDUInfo.table.numCols; i++ ) {
	    Tcl_AppendElement(curFile->interp,
			      curFile->CHDUInfo.table.colName[i]);
	 }

      } else {
	       
	 /*******************************************
	  *  Return info about one or more columns  *
	  *******************************************/

	 if( !strcmp(argv[3], "-stat") ) {

	    if ( argc < 5 ) {
	       Tcl_SetResult(curFile->interp,
			     "Usage: info column -stat columnName ?felem? ?rows?",
			     TCL_STATIC);
	       return TCL_ERROR;
	    }
	    
	    if( argc == 5 ) { 
	       felem = 1;
	    } else if( Tcl_GetInt(curFile->interp, argv[5], &felem)
		       != TCL_OK ) {
	       return TCL_ERROR;
	    }
	       
	    if( argc >= 7 ) {
	       if( fitsParseRange(argv[6],&numRange,range,FITS_MAXRANGE,
				  1, curFile->CHDUInfo.table.numRows,errMsg)
		   != TCL_OK ) {
		  Tcl_SetResult(curFile->interp,
				"Error parsing row range:\n", TCL_STATIC);
		  Tcl_AppendResult(curFile->interp, errMsg, (char*)NULL);
		  return TCL_ERROR;
	       }
	    } else {
	       numRange = 1;
	       range[0][0] = 1;
	       range[0][1] = curFile->CHDUInfo.table.numRows ;
	    }    
	       
	    if( fitsTransColList( curFile, argv[4], &numCols,
				  colNums, colTypes, strSize) != TCL_OK )
	       return TCL_ERROR;	      
	    
	    if( fitsColumnStatistics(curFile,colNums[0],felem,
				     numRange,range) != TCL_OK ) {
	       return TCL_ERROR;
	    }
	    
	 } else if( !strcmp(argv[3], "-minmax") ) {

	    if ( argc < 5 ) {
	       Tcl_SetResult(curFile->interp,
			     "Usage: info column -minmax "
			     "columnName ?felem? ?rows?", TCL_STATIC);
	       return TCL_ERROR;
	    }
	    
	    if( argc == 5 ) { 
	       felem = 1;
	    } else if( Tcl_GetInt(curFile->interp, argv[5], &felem)
		       != TCL_OK ) {
	       return TCL_ERROR;
	    } 
	    
	    if( argc >= 7 ) {
	       if( fitsParseRange(argv[6],&numRange,range,FITS_MAXRANGE,
				  1, curFile->CHDUInfo.table.numRows,errMsg)
		   != TCL_OK ) {
		  Tcl_SetResult(curFile->interp,
				"Error parsing row range:\n", TCL_STATIC);
		  Tcl_AppendResult(curFile->interp, errMsg, (char*)NULL);
		  return TCL_ERROR;
	       }
	    } else {
	       numRange = 1;
	       range[0][0] = 1;
	       range[0][1] = curFile->CHDUInfo.table.numRows ;
	    }    
	    
	    for ( i = 0; i < curFile->CHDUInfo.table.numCols; i++) {
	       if( !strcasecmp(argv[4],curFile->CHDUInfo.table.colName[i]) ) {
		  if( fitsColumnMinMax(curFile, i+1, felem, numRange, range)
		      != TCL_OK ) {
		     return TCL_ERROR;
		  }
		  break;
	       }
	    }
	    
	 } else if( !strcmp(argv[3], "-exact") ) {

	    /*************************************************************
	     *  Return Info about Columns matching an exact column name  *
	     *************************************************************/

	    if( argc != 5 ) {
	       Tcl_SetResult(curFile->interp,
			     "Usage: info column -exact columnNames",
			     TCL_STATIC);
	       return TCL_ERROR;
	    }
	    if( fitsTransColList( curFile, argv[4], &numCols,
				  colNums, colTypes, strSize) != TCL_OK )
	       return TCL_ERROR;
	    
	    for ( i = 0; i < numCols; i++ ) {
	       j = colNums[i]-1;
	       mrgList[0] = curFile->CHDUInfo.table.colName[j];
	       mrgList[1] = curFile->CHDUInfo.table.colType[j];
	       mrgList[2] = curFile->CHDUInfo.table.colUnit[j];
	       mrgList[3] = curFile->CHDUInfo.table.colDisp[j];
	       mrgList[4] = curFile->CHDUInfo.table.colFormat[j];
	       sprintf(tmpStr[0],  "%d",
		       curFile->CHDUInfo.table.colWidth[j]);
	       mrgList[5] = tmpStr[0];
	       sprintf(tmpStr[1], "%d",
		       curFile->CHDUInfo.table.colTzflag[j]);
	       mrgList[6] = tmpStr[1];
	       sprintf(tmpStr[2], "%d",
		       curFile->CHDUInfo.table.colTsflag[j]);
	       mrgList[7] = tmpStr[2];
	       mrgList[8] = curFile->CHDUInfo.table.colNull[j];
	       Tcl_AppendElement(curFile->interp,Tcl_Merge(9,mrgList));
	    }
	    
	 } else if( argc==4 ) {

	    /***********************************************************
	     *  Return Info about Columns matching regular expression  *
	     ***********************************************************/

	    Tcl_DStringInit(&concatList);

	    if( Tcl_SplitList(curFile->interp, argv[3], &numCols,
			      &colList) != TCL_OK ) {
	       return TCL_ERROR;
	    }

	    if( fitsMakeRegExp(curFile->interp, numCols, colList,
			       &concatList, 1)
		== TCL_ERROR ) {
	       Tcl_SetResult(curFile->interp,
			     "Error making up reg expr", TCL_STATIC);
	       Tcl_DStringFree(&concatList);
	       ckfree((char*)colList);
	       return TCL_ERROR;
	    }
	    ckfree((char*)colList);
	    pattern = Tcl_DStringValue(&concatList);
	    for ( i = 0; i < curFile->CHDUInfo.table.numCols; i++) {
	       strToUpper(curFile->CHDUInfo.table.colName[i], &tmpStrPtr);
	       status = Tcl_RegExpMatch(curFile->interp, tmpStrPtr, pattern);
	       ckfree( (char*)tmpStrPtr );
	       if( status == 1 ) {
		  mrgList[0] = curFile->CHDUInfo.table.colName[i];
		  mrgList[1] = curFile->CHDUInfo.table.colType[i];
		  mrgList[2] = curFile->CHDUInfo.table.colUnit[i];
		  mrgList[3] = curFile->CHDUInfo.table.colDisp[i];
		  mrgList[4] = curFile->CHDUInfo.table.colFormat[i];
		  sprintf(tmpStr[0],  "%d",
			  curFile->CHDUInfo.table.colWidth[i]);
		  mrgList[5] = tmpStr[0];
		  sprintf(tmpStr[1], "%d",
			  curFile->CHDUInfo.table.colTzflag[i]);
		  mrgList[6] = tmpStr[1];
		  sprintf(tmpStr[2], "%d",
			  curFile->CHDUInfo.table.colTsflag[i]);
		  mrgList[7] = tmpStr[2];
		  mrgList[8] = curFile->CHDUInfo.table.colNull[i];
		  Tcl_AppendElement(curFile->interp,Tcl_Merge(9,mrgList));
	       } else if( status == -1 ) {
		  Tcl_AppendResult(curFile->interp,"Error, ", pattern,
				   " not a Regular Expression.",
				   (char *) NULL);
		  Tcl_DStringFree(&concatList);
		  return TCL_ERROR;
	       }
	    }
	    Tcl_DStringFree(&concatList);

	 } else {

	    Tcl_SetResult(curFile->interp,
			  "Usage:\n"
			  "      info column ?-exact? colNames \n"
			  "                  -minmax  colName firstElement ?rowRange? \n"
			  "                  -stat    colName firstElement ?rowRange? \n",
			  TCL_STATIC);
	    return TCL_ERROR;

	 }
      }

      /*  End of 'info column'  */

   } else if( !strcmp("expr", argv[2]) ) {
      
      if( curFile->hduType == IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,"Not a table extension", TCL_STATIC);
	 return TCL_ERROR;
      }
      if( argc != 4 ) {
	 Tcl_SetResult(curFile->interp, 
		       "Usage: info expr exprStr", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( exprGetInfo( curFile, argv[3] ) ) {
	 return TCL_ERROR;
      }

   } else if( !strcmp("imgdim", argv[2]) ) {
      
      if ( curFile->hduType != IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not an image", TCL_STATIC);
	 return TCL_ERROR;
      }

      Tcl_ResetResult(curFile->interp);
      for (i=0; i < curFile->CHDUInfo.image.naxes; i++) {
	 sprintf(tmpStr[0], "%ld", curFile->CHDUInfo.image.naxisn[i]);
	 Tcl_AppendElement(curFile->interp, tmpStr[0]);
      }

   } else {

      Tcl_SetResult(curFile->interp,
		    "Unrecognized option to info", TCL_STATIC);
      return TCL_ERROR;
      
   }

   return TCL_OK;
}


/******************************************************************
 *                             Get
 ******************************************************************/

int fitsTcl_get( FitsFD *curFile, int argc, char *argv[] )
{
   static char *getList = "\n" 
      "Available Commands:\n"
      "get keyword ?keyName?   - displays the keyword(s) keyname\n"
      "                        - keywords are specified by reg. expression\n"
      "get keyword -num keyNum - displays the num th keyword in the CHDU\n"
      "get wcs ?RAcol DECcol?\n"
      "    - return a list of the WCS parameters for either a table or image:\n" 
      "        {xrval yrval xrpix yrpix xinc yinc rot ctype}\n" 
      "      For a table, supply RAcol and DECcol which are column names or \n"
      "      numbers of the RA column and DEC column\n"
      "get image ?firstElem? ?numElem?\n"
      "                        - return elements of an image\n"
      "get table ?-c? ?colList? ?rowList?\n"
      "                        - return the elements rowList from list colList\n"
      "                        - if no rowList is provided, give all rows\n"
      "                        - if no colList is provided, give all columns\n"
      "                        - use colList = * for all columns\n"
      "                        - -c means return each column as a seperate list.\n"
      "get vtable colname firstelement\n"
      "                        - get the firstelement-th vector element\n"
      "\n";
   
   int numCols,colNums[FITS_COLMAX],colTypes[FITS_COLMAX],strSize[FITS_COLMAX];
   int numRange, range[FITS_MAXRANGE][2];
   Tcl_HashEntry *newEntry;
   Tcl_HashSearch search;
   Tcl_DString concatList, regExpList;
   Tcl_DString ** colDString;
   FitsCardList *curCard;
   Keyword *newKwd;
   char errMsg[256];
   int nmove,i,k,l;
   int bycol,niters,fRow;
   int ntodo,felem;
   char ***strValArray;
   char *pattern;
   int status = 0;

   if ( argc == 2 ) {
      Tcl_SetResult(curFile->interp, getList, TCL_STATIC);
      return TCL_OK;
   }
   
   if( !strcmp("keyword", argv[2]) ) {

      /* GET KEYWORD */

      if( argc == 3 ) {
	 
	 Tcl_DStringInit(&concatList);
	 
	 newEntry = Tcl_FirstHashEntry(curFile->kwds,&search);
	 while ( newEntry ) {
	    newKwd = (Keyword *) Tcl_GetHashValue(newEntry);
	    Tcl_DStringStartSublist(&concatList);
	    Tcl_DStringAppendElement(&concatList,newKwd->name);
	    Tcl_DStringAppendElement(&concatList,newKwd->value);
	    Tcl_DStringAppendElement(&concatList,newKwd->comment);
	    Tcl_DStringEndSublist(&concatList);
	    newEntry = Tcl_NextHashEntry(&search);
	 }
	 Tcl_DStringResult(curFile->interp,&concatList);

      } else if( !strcmp(argv[3],"-num") ) {

	 if ( 5 != argc ) {
	    Tcl_SetResult(curFile->interp,
			  "Wrong number of args, expected get keyword "
			  "-num number", TCL_STATIC);
	    return TCL_ERROR;
	 }

	 if( Tcl_GetInt(curFile->interp,argv[4],&nmove) != TCL_OK ) {
	    Tcl_AppendResult(curFile->interp,
			     "\nWrong type for nmove",(char *) NULL);
	    return TCL_ERROR;
	 }
	 
	 /*
	  * First look through the comments and the history cards:
	  * remember the first card is always a dummy...
	  */

	 curCard = (curFile->hisHead)->next;
	 while( curCard ) {
	    if ( curCard->pos == nmove ) {
	       Tcl_AppendElement(curFile->interp,"HISTORY");
	       Tcl_AppendElement(curFile->interp," ");
	       Tcl_AppendElement(curFile->interp,curCard->value);
	       return TCL_OK;
	    }
	    curCard = curCard->next;
	 }
	 
	 curCard = (curFile->comHead)->next;
	 while( curCard ) {
	    if ( curCard->pos == nmove ) {
	       Tcl_AppendElement(curFile->interp,"COMMENT");
	       Tcl_AppendElement(curFile->interp," ");
	       Tcl_AppendElement(curFile->interp,curCard->value);
	       return TCL_OK;
	    }
	    curCard = curCard->next;
	 }
	 
	 newEntry = Tcl_FirstHashEntry(curFile->kwds,&search);
	 while( newEntry ) {
	    newKwd = (Keyword *) Tcl_GetHashValue(newEntry);
	    if ( newKwd->pos == nmove ) {
	       Tcl_AppendElement(curFile->interp,newKwd->name);
	       Tcl_AppendElement(curFile->interp,newKwd->value);
	       Tcl_AppendElement(curFile->interp,newKwd->comment);
	       return TCL_OK;
	    }
	    newEntry = Tcl_NextHashEntry(&search);
	 }
	 sprintf(errMsg,"Keyword number %d not found",nmove);
	 Tcl_SetResult(curFile->interp,errMsg,TCL_VOLATILE);
	 return TCL_ERROR;

      } else {
	 
	 Tcl_DStringInit(&regExpList);
	 
	 if( fitsMakeRegExp(curFile->interp, argc-3, argv+3, &regExpList, 1)
	     == TCL_ERROR ) {
	    Tcl_SetResult(curFile->interp,
			  "Error building regular expression", TCL_STATIC);
	    Tcl_DStringFree(&regExpList);
	    return TCL_ERROR;
	 }
	 
	 pattern = Tcl_DStringValue(&regExpList);
	 
	 Tcl_DStringInit(&concatList);
	 
	 niters = 0;
	 newEntry = Tcl_FirstHashEntry(curFile->kwds,&search);
	 while ( NULL != newEntry ) {
	    newKwd = (Keyword *) Tcl_GetHashValue(newEntry);
	    status = Tcl_RegExpMatch(curFile->interp,newKwd->name,pattern);
	    if ( status == 1 ) {
	       niters = 1;
	       Tcl_DStringStartSublist(&concatList);
	       Tcl_DStringAppendElement(&concatList,newKwd->name);
	       Tcl_DStringAppendElement(&concatList,newKwd->value);
	       Tcl_DStringAppendElement(&concatList,newKwd->comment);
	       Tcl_DStringEndSublist(&concatList);
	       newEntry = Tcl_NextHashEntry(&search);
	    } else if ( status == -1 ) {
	       Tcl_AppendResult(curFile->interp,"The Pattern: ",pattern,
				" is not a regular expression."
				,(char *) NULL);
	       Tcl_DStringFree(&concatList);
	       Tcl_DStringFree(&regExpList);
	       return TCL_ERROR;
	    } else {
	       newEntry = Tcl_NextHashEntry(&search);
	    }
	 }
	 if( !niters ) {
	    Tcl_SetResult(curFile->interp,
			  "No matching keywords found/or keyword not loaded",
			  TCL_STATIC);
	    Tcl_DStringFree(&concatList);
	    return TCL_ERROR;
	 }

	 Tcl_DStringResult(curFile->interp,&concatList);
      }	

   } else if( !strcmp("wcs", argv[2]) ) {

      /*  Get WCS  */

      if ( curFile->hduType == IMAGE_HDU ) {

	 /*  Get WCS from Image extension  */

	 if( argc != 3 ) {
	    Tcl_SetResult(curFile->interp,
			  "For image extension use, get wcs", TCL_STATIC);
	    return TCL_ERROR;
	 }

	 if( fitsImageGetWcs(curFile) != TCL_OK ) {
	    return TCL_ERROR;
	 }

      } else {

	 /*  Get WCS from Table extension  */

	 int ranum = 0;
	 int decnum = 0;

	 if( argc != 5 ) {
	    Tcl_SetResult(curFile->interp,
			  "For table extension use, get wcs RAcol DecCol",
			  TCL_STATIC);
	    return TCL_ERROR;
	 }

	 if( Tcl_GetInt(curFile->interp, argv[3], &ranum) != TCL_OK ) {
	    Tcl_ResetResult(curFile->interp);
	    if( fitsTransColList( curFile, argv[3],
				  &numCols, colNums, colTypes, strSize)
		!= TCL_OK ) {
	       Tcl_SetResult(curFile->interp,
			     "Unable to read RAcol", TCL_STATIC);
	       return TCL_ERROR;
	    }
	    if( numCols != 1 ) {
	       Tcl_SetResult(curFile->interp,
			     "Can only have 1 RAcol value", TCL_STATIC);
	       return TCL_ERROR;
	    }
	    ranum = colNums[0];
	 }

	 if( Tcl_GetInt(curFile->interp, argv[4], &decnum) != TCL_OK ) {
	    Tcl_ResetResult(curFile->interp);
	    if( fitsTransColList( curFile, argv[4],
				  &numCols, colNums, colTypes, strSize)
		!= TCL_OK ) {
	       Tcl_SetResult(curFile->interp,
			     "Unable to read DecCol", TCL_STATIC);
	       return TCL_ERROR;
	    }
	    if( numCols != 1 ) {
	       Tcl_SetResult(curFile->interp,
			     "Can only have 1 DecCol value", TCL_STATIC);
	       return TCL_ERROR;
	    }
	    decnum = colNums[0];
	 }

	 if( fitsTableGetWcs(curFile, ranum, decnum) != TCL_OK ) {
	    return TCL_ERROR;
	 }
      
      }

   } else if( !strcmp("imgwcs", argv[2]) ) {

      /*  Get IMGWCS  */

      if ( curFile->hduType != IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not an image", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( fitsImageGetWcs(curFile) != TCL_OK ) {
	 return TCL_ERROR;
      }

   } else if( !strcmp("colwcs", argv[2]) ) {

      /*  Get COLWCS  */

      int ranum = 0;
      int decnum = 0;

      if( curFile->hduType == IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not a table", TCL_STATIC);
	 return TCL_ERROR;
      }
      if( argc != 5 ) {
	 Tcl_SetResult(curFile->interp,
		       "get colwcs RAcol DECcol", TCL_STATIC);
	 return TCL_ERROR;
      }

      if( Tcl_GetInt(curFile->interp, argv[3], &ranum) != TCL_OK ) {
	 Tcl_ResetResult(curFile->interp);
	 if( fitsTransColList( curFile, argv[3],
			       &numCols, colNums, colTypes, strSize)
	     != TCL_OK ) {
	    Tcl_SetResult(curFile->interp,
			  "Unable to read RAcol", TCL_STATIC);
	    return TCL_ERROR;
	 }
	 if( numCols != 1 ) {
	    Tcl_SetResult(curFile->interp,
			  "Can only have 1 RAcol value", TCL_STATIC);
	    return TCL_ERROR;
	 }
	 ranum = colNums[0];
      }

      if( Tcl_GetInt(curFile->interp, argv[4], &decnum) != TCL_OK ) {
	 Tcl_ResetResult(curFile->interp);
	 if( fitsTransColList( curFile, argv[4],
			       &numCols, colNums, colTypes, strSize)
	     != TCL_OK ) {
	    Tcl_SetResult(curFile->interp,
			  "Unable to read DecCol", TCL_STATIC);
	    return TCL_ERROR;
	 }
	 if( numCols != 1 ) {
	    Tcl_SetResult(curFile->interp,
			  "Can only have 1 DecCol value", TCL_STATIC);
	    return TCL_ERROR;
	 }
	 decnum = colNums[0];
      }

      if( fitsTableGetWcsOld(curFile, ranum, decnum) != TCL_OK ) {
	 return TCL_ERROR;
      }
      
   } else if( !strcmp("image", argv[2]) ) {

      long fElem, nElem;

      if( argc < 3 || argc > 5 ) {
	 Tcl_SetResult(curFile->interp,
		       "get image firstElem numElem", TCL_STATIC);
	 return TCL_ERROR;
      }

      if( curFile->hduType != IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not a table", TCL_STATIC);
	 return TCL_ERROR;
      }

      if( argc>3 ) {
	 fElem = atol( argv[3] );
	 if( argc>4 ) {
	    nElem = atol( argv[4] );
	 } else {
	    nElem = 1;
	 }
      } else {
	 fElem = 1;
	 nElem = 1;
	 i = curFile->CHDUInfo.image.naxes;
	 while( i-- )
	    nElem *= curFile->CHDUInfo.image.naxisn[i];
      }

      if( imageBlockLoad_1D(curFile, fElem, nElem) != TCL_OK ) {
	 return TCL_ERROR;
      }

   } else if ( !strcmp("imageblock", argv[2]) ) {

      /*  GET IMAGE in blocks  */

      long slice = 1;

      if( argc < 8 || argc > 9 ) {
	 Tcl_SetResult(curFile->interp,
		       "FitsHandle get imageblock arrayName firstRow "
		       "numRows firstCol numCols ?slice?", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( curFile->hduType != IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not a table", TCL_STATIC);
	 return TCL_ERROR;
      }

      if( argc > 8 ) 
	 slice = atol(argv[8]);
      
      if( imageBlockLoad(curFile, argv[3], atol(argv[4]), atol(argv[5]),
			 atol(argv[6]), atol(argv[7]), slice )
	  != TCL_OK ) {
	 return TCL_ERROR;  /*  Sets own error message  */
      }

   } else if( !strcmp("table",argv[2] ) ) {

      /* GET TABLE */

      if ( curFile->hduType == IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not a table", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( curFile->CHDUInfo.table.loadStatus != 1 ) {
	 Tcl_SetResult(curFile->interp,
		       "Need to load the hdu first", TCL_STATIC);
	 return TCL_ERROR;
      }

      /*
       * Strip off the "-c" flag if present... 
       */
      
      if( 3 < argc && !strcmp(argv[3],"-c") ) {
	 bycol = 1;
	 for ( i = 3; i < argc - 1; i++ ) {
	    argv[i] = argv[i+1];
	 }
	 argc--;
      } else {
	 bycol = 0;
      }
      
      if( argc > 5 ) {
	 Tcl_SetResult(curFile->interp,
		       "Wrong number of arguments, need "
		       "'get table ?-c? ?columns? ?rows?'", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      /* If no colList is given, or it is "*", use all the columns... */ 
      
      if( TCL_OK !=
	  fitsTransColList( curFile, ( argc==3 ? "*" : argv[3] ),
			    &numCols, colNums, colTypes, strSize) )
	 return TCL_ERROR;
      
      /* 
       * Get the Row range parameter 
       */
      
      if( argc < 5 ) { 
	 numRange    = 1;
	 range[0][0] = 1;
	 range[0][1] = curFile->CHDUInfo.table.numRows;
      } else {
	 if( fitsParseRange(argv[4],&numRange,range,FITS_MAXRANGE,
			    1, curFile->CHDUInfo.table.numRows,errMsg)
	     != TCL_OK ) {
	    Tcl_SetResult(curFile->interp,
			  "Error parsing row range:\n", TCL_STATIC);
	    Tcl_AppendResult(curFile->interp, errMsg, (char*)NULL);
	    return TCL_ERROR;
	 }
      }
      
      /*
       *  Now allocate an array for the results.
       */

      strValArray = (char ***) ckalloc(numCols * sizeof(char ***));
      for( i = 0; i < numCols; i++) {
	 strValArray[i] = (char **) makeContigArray(FITS_CHUNKSIZE,
						    strSize[i] + 1,'c');
      }

      /* Now get the rows... */
      
      if ( bycol ) {
	 colDString = (Tcl_DString ** ) ckalloc ( numCols 
						  * sizeof(Tcl_DString *));
	 for(i = 0; i < numCols; i++ ) {
	    colDString[i] = (Tcl_DString *) ckalloc(sizeof(Tcl_DString));
	    Tcl_DStringInit(colDString[i]);
	 }
	 Tcl_DStringStartSublist(colDString[0]);
      } else {
	 Tcl_DStringInit(&concatList);
      }
      
      for (i = 0; i < numRange; i++ ) {
	 fRow  = range[i][0];
	 while( fRow <= range[i][1] ) {
	    ntodo  = range[i][1] - fRow + 1;
	    if( ntodo>FITS_CHUNKSIZE ) ntodo = FITS_CHUNKSIZE;
	    status = tableGetToString(curFile, fRow, 1, ntodo,
				      numCols, colTypes, colNums,
				      strSize, strValArray);
	    if( status != TCL_OK )
	       break;
	    fRow += ntodo;
	    if ( bycol ) {
	       for( k = 0; k < numCols; k++) {
		  for ( l = 0; l < ntodo; l++) {
		     Tcl_DStringAppendElement(colDString[k],
					      strValArray[k][l]);
		  }
	       }
	    } else {
	       for ( l = 0; l < ntodo; l++) {
		  Tcl_DStringStartSublist(&concatList);
		  for( k = 0; k < numCols; k++) {
		     Tcl_DStringAppendElement(&concatList,
					      strValArray[k][l]);
		  }
		  Tcl_DStringEndSublist(&concatList);
	       }
	    }		    
	 }
      } 
      
      if( status ) {

	 if ( bycol ) {
	    for ( i = 0; i < numCols ; i++ ) {  
	       Tcl_DStringFree(colDString[i]);
	    }
	 } else {
	    Tcl_DStringFree(&concatList);
	 }
	 
      } else {

	 if ( bycol ) {
	    Tcl_DStringEndSublist(colDString[0]);
	    for ( i = 1; i < numCols; i++ ) {
	       Tcl_DStringAppendElement(colDString[0],
					Tcl_DStringValue(colDString[i]));
	       Tcl_DStringFree(colDString[i]);
	       ckfree((char *) colDString[i]);
	    }
	    Tcl_DStringResult(curFile->interp,colDString[0]);
	    ckfree((char *) colDString);
	 } else {
	    Tcl_DStringResult(curFile->interp,&concatList);
	 }

      }
      
      for ( i = 1; i < numCols; i++ ) {
	 ckfree((char *) strValArray[i][0]);
	 ckfree((char *) strValArray[i]);
      }
      ckfree((char *) strValArray);
      
      if( status ) return TCL_ERROR;
      
   } else if( !strcmp("vtable",argv[2]) ) {

      /* GET vector from the TABLE */

      if( curFile->hduType == IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not a table", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( curFile->CHDUInfo.table.loadStatus != 1 ){
	 Tcl_SetResult(curFile->interp,
		       "Need to load the hdu first", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( argc != 5 ) {
	 Tcl_SetResult(curFile->interp,
		       "Wrong number of arguments, need "
		       "'get vtable column felem'", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      
      if( fitsTransColList( curFile, argv[3],
			    &numCols, colNums, colTypes, strSize )
	  != TCL_OK )
	 return TCL_ERROR;

      if( numCols != 1 ) {
	 Tcl_SetResult(curFile->interp,
		       "Can only read one vector column of a table at a time",
		       TCL_STATIC);
      }

      numRange    = 1;
      range[0][0] = 1;
      range[0][1] = curFile->CHDUInfo.table.numRows;

      /*
       * Now allocate an array for the results.  It has to be long enough to 
       * write in the doubles we want...:
       */

      strValArray = (char ***) ckalloc( 1 * sizeof(char ***));
      strValArray[0] = (char **) makeContigArray(FITS_CHUNKSIZE,
						 strSize[0] + 1, 'c');

      /* Now get the rows... */
      
      Tcl_DStringInit(&concatList);

      felem = atoi(argv[4]);
      fRow  = range[0][0];
      while( fRow <= range[0][1] ) {
	 ntodo = range[0][1] - fRow + 1;
	 if( ntodo>FITS_CHUNKSIZE ) ntodo = FITS_CHUNKSIZE;
	 status = tableGetToString(curFile, fRow, felem, ntodo,
				   numCols, colTypes, colNums,
				   strSize, strValArray);
	 if( status != TCL_OK )
	    break;
	 fRow += ntodo;
	 for ( l = 0; l < ntodo; l++) {
	    Tcl_DStringAppendElement(&concatList,
				     strValArray[0][l]);
	 }
      }
      
      if( status ) {
	 Tcl_DStringFree(&concatList);
      } else {
	 Tcl_DStringResult(curFile->interp, &concatList);
      }
      
      ckfree((char *) strValArray[0][0]);
      ckfree((char *) strValArray[0]);
      ckfree((char *) strValArray);
      
      if( status ) return TCL_ERROR;
      
   } else {

      Tcl_SetResult(curFile->interp,
		    "ERROR: unrecognized command to get", TCL_STATIC);
      return TCL_ERROR;

   }

   return TCL_OK;
}


/******************************************************************
 *                             Put
 ******************************************************************/

int fitsTcl_put( FitsFD *curFile, int argc, char *argv[] )
{
   static char *putKeyList = "put keyword ?-num n? card ?formatFlag?";
   static char *putHisList = "put history string";
   static char *putTabList =
      "put table colName firstElem rowSpan listOfData\n";
   
   static char *putImgList = "put image firstElem numElem listOfData\n";
   
   static char *putIhdList =
      "put ihd ?-p? ?bitpix naxis naxesList? \n"
      "             - -p primary extension \n";
   
   static char *putAhdList = 
      "put ahd numRows numCols {colName} {colType} {colUnit} {tbCol}\n"
      "                                            extname rowLength\n"
      "       - colType: L(logical), X(bit), I(16 bit integer), "
                                                  "J(32 bit integer)\n"
      "                  An(n Character), En(Single with n format), \n"
      "                  Dn(Double with n format), B(Unsigned) \n"
      "                  C(Complex), M(Double complex)  ";
   
   static char *putBhdList = 
      "put bhd numRows numCols {colName} {colType} {colUnit} extname \n"
      "       - colType: nL(logical),nX(bit), nI(16 bit integer), "
                                                   "nJ(32 bit integer)\n"
      "                  nA(Character), nE(Single), nD(Double), nB(Unsigned) \n"
      "                  nC(Complex), M(Double complex)  ";
   
   int numCols,colNums[FITS_COLMAX],colTypes[FITS_COLMAX],strSize[FITS_COLMAX];
   int numRange, range[FITS_MAXRANGE][2];
   char errMsg[256];
   int i;

   if ( argc == 2 ) {
      Tcl_SetResult(curFile->interp,"Available Commands:\n",TCL_STATIC);
      Tcl_AppendResult(curFile->interp, putKeyList,"\n", (char *)NULL);
      Tcl_AppendResult(curFile->interp, putTabList,"\n", (char *)NULL);
      Tcl_AppendResult(curFile->interp, putIhdList,"\n", (char *)NULL);
      Tcl_AppendResult(curFile->interp, putAhdList,"\n", (char *)NULL);
      Tcl_AppendResult(curFile->interp, putBhdList,"\n", (char *)NULL);
      return TCL_OK;
   }

   if( !strcmp( "keyword", argv[2] ) ) {

      /* Write Keyword */

      int format, cardNum=0, recLoc=3;

      if( argc < 4 || argc > 7 ) {
	 Tcl_SetResult(curFile->interp, putKeyList, TCL_STATIC);
	 return TCL_ERROR;
      }

      if( !strcmp(argv[3], "-num") ) {
	 if( argc < 6 ) {
	    Tcl_SetResult(curFile->interp, putKeyList, TCL_STATIC);
	    return TCL_ERROR;
	 }
	 if( Tcl_GetInt(curFile->interp, argv[4], &cardNum) != TCL_OK ) {
	    return TCL_ERROR;
	 }
	 recLoc += 2;
      }

      if( recLoc+1 < argc ) {
	 if( Tcl_GetInt(curFile->interp, argv[recLoc+1], &format) != TCL_OK ) {
	    return TCL_ERROR;
	 }
      } else {
	 format = 1;
      }

      if( fitsPutKwds(curFile, cardNum, argv[recLoc], format)
	  != TCL_OK ) {
	 return TCL_ERROR;
      } 	  

   } else if( !strcmp( "history", argv[2] ) ) {

      /*  Write History  */

      if( argc != 4 ) {
	 Tcl_SetResult(curFile->interp, putHisList, TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( fitsPutHisKwd(curFile, argv[3]) != TCL_OK ) {
	 return TCL_ERROR;
      }
      
   } else if( !strcmp ( "image", argv[2] ) ) {

      /*  Write Image  */

      if( curFile->hduType != IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not an image", TCL_STATIC);
	 return TCL_ERROR;
      }
      if( argc != 6 ) {
	 Tcl_SetResult(curFile->interp, putImgList, TCL_STATIC);
	 return TCL_ERROR;
      }
      if( varSaveToImage(curFile, 
			 atol(argv[3]),
			 atol(argv[4]), 
			 argv[5]) != TCL_OK ) {
	 return TCL_ERROR;
      }

   } else if( !strcmp( "table", argv[2] ) ) {
      
      /*  Write Table  */ 

      if ( curFile->hduType == IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not a table", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if ( argc != 7 ) {
	 Tcl_SetResult(curFile->interp, putTabList, TCL_STATIC);
	 return TCL_ERROR;
      }

      /* parse the column name */

      if( fitsTransColList(curFile,argv[3],
			   &numCols,colNums,colTypes,strSize) != TCL_OK ) {
	 return TCL_ERROR;
      }
      if( numCols != 1 ) {
	 Tcl_SetResult(curFile->interp,
		       "Can only write one column at a time", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      /* 
       * Get the Row range parameter 
       */

      if( fitsParseRange(argv[5],&numRange,range,FITS_MAXRANGE,
			 1, curFile->CHDUInfo.table.numRows,errMsg)
	  != TCL_OK ) {
	 Tcl_SetResult(curFile->interp,
		       "Error parsing row range:\n", TCL_STATIC);
	 Tcl_AppendResult(curFile->interp, errMsg, (char*)NULL);
	 return TCL_ERROR;
      }     
      if( numRange != 1 ) {
	 Tcl_SetResult(curFile->interp,
		       "Can only write one row range at a time", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( varSaveToTable(curFile, 
			 colNums[0], 
			 range[0][0], 
			 atol(argv[4]),           /* first Element*/
			 range[0][1]-range[0][0]+1,
			 argv[6]) != TCL_OK ) {
	 return TCL_ERROR;
      }
      
   } else if( !strcmp( "ihd", argv[2] ) ) {

      /*  Write Image Header  */

      int isPrimary;
      if ( argc < 4 || argc > 7 ) {
	 Tcl_SetResult(curFile->interp, putIhdList, TCL_STATIC);
	 return TCL_ERROR;
      }

      /*
       *  Strip off the "-p" flag if present... 
       */
      
      if( !strcmp(argv[3],"-p") ) {
	 isPrimary = 1;
	 for ( i = 3; i < argc - 1; i++ ) {
	    argv[i] = argv[i+1];
	 }
	 argc--;
      } else {
	 isPrimary = 0;
      }
      
      if( fitsPutReqKwds(curFile, isPrimary, IMAGE_HDU, argc, argv) 
	  !=TCL_OK ) {
	 return TCL_ERROR;
      }

   } else if( !strcmp( "ahd", argv[2] ) ) {

      /*  Write ASCII Table Header  */

      if( argc != 11 ) {
	 Tcl_SetResult(curFile->interp, putAhdList, TCL_STATIC);
	 return TCL_ERROR;
      }
      
      /*  Strip off the numCols parameter... use colNames length instead  */

      for ( i = 4; i < argc - 1; i++ ) {
	 argv[i] = argv[i+1];
      }
      argc--;

      if( fitsPutReqKwds(curFile, 0, ASCII_TBL, argc, argv)
	  != TCL_OK ) {
	 return TCL_ERROR;
      }
      
   } else if( !strcmp( "bhd", argv[2] ) ) {

      /*  Write Binary Table Header  */

      if( argc != 9 ) {
	 Tcl_SetResult(curFile->interp, putBhdList, TCL_STATIC);
	 return TCL_ERROR;
      }
      
      /*  Strip off the numCols parameter... use colNames length instead  */

      for ( i = 4; i < argc - 1; i++ ) {
	 argv[i] = argv[i+1];
      }
      argc--;

      if( fitsPutReqKwds(curFile, 0, BINARY_TBL, argc, argv)
	  != TCL_OK ) {
	 return TCL_ERROR;
      }
      
   } else {
   
      Tcl_SetResult(curFile->interp, "Unknown put function", TCL_STATIC);
      return TCL_ERROR;

   }

   return TCL_OK;
}


/******************************************************************
 *                             Insert
 ******************************************************************/

int fitsTcl_insert( FitsFD *curFile, int argc, char *argv[] )
{
   static char *insertList[] = {
      "insert keyword index record ?formatflag?",
      "insert column  index colName colForm",
      "insert row     index numRows",
      "insert image ?-p? ?bitpix naxis naxesList? \n"
      "             - -p primary extension, keywords optional if empty array",
      "insert table numRows {colNames} {colForms} ?{colUnits} extname?\n"
      "       - colForm: nL(logical),nX(bit), nI(16 bit integer), "
                                                   "nJ(32 bit integer)\n"
      "                  nA(Character), nE(Single), nD(Double), nB(Unsigned) \n"
      "                  nC(Complex), M(Double complex) \n"
      "insert table -ascii numRows {colNames} {colForms} ?{colUnits}\n"
      "                                            {tbCols} extname rowWidth?\n"
      "       - colForm: L(logical), X(bit), I(16 bit integer), "
                                                  "J(32 bit integer)\n"
      "                  An(n Character), En(Single with n format), \n"
      "                  Dn(Double with n format), B(Unsigned) \n"
      "                  C(Complex), M(Double complex)  " };
   
   int index, format, numRows, i;

   if( argc == 2 ) {
      Tcl_AppendResult(curFile->interp,
		       "Available commands:\n",
		       insertList[0], "\n",
		       insertList[1], "\n",
		       insertList[2], "\n",
		       insertList[3], "\n",
		       insertList[4], "\n",
		       (char *)NULL);
      return TCL_ERROR;
   }
   
   if( !strcmp( "keyword", argv[2] ) ) { 

      if( argc < 5 || argc > 6 ) {
	 Tcl_SetResult(curFile->interp, insertList[0], TCL_STATIC);
	 return TCL_OK;
      }
      
      if( Tcl_GetInt(curFile->interp, argv[3], &index) != TCL_OK) {
	 Tcl_SetResult(curFile->interp,
		       "Failed to get integer index", TCL_STATIC);
	 return TCL_ERROR;
      }

      if( argc==6 ) {
	 if ( Tcl_GetInt(curFile->interp, argv[5], &format) != TCL_OK) {
	    Tcl_SetResult(curFile->interp,
			  "Failed to get integer format flag", TCL_STATIC);
	    return TCL_ERROR;
	 }
      } else {
	 format = 1;
      }
      
      if( fitsInsertKwds(curFile, index, argv[4], format) != TCL_OK ) {
	 return TCL_ERROR;
      } 

   } else if( !strcmp( "column", argv[2] ) ) {

      if (argc != 6 ) {
	 Tcl_SetResult(curFile->interp, insertList[1], TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( Tcl_GetInt(curFile->interp, argv[3], &index) != TCL_OK) {
	 Tcl_SetResult(curFile->interp,
		       "Failed to get integer index", TCL_STATIC);
	 return TCL_ERROR;
      }

      if( addColToTable(curFile,index,argv[4],argv[5]) != TCL_OK ) {
	 return TCL_ERROR;
      }       

   } else if( !strcmp( "row", argv[2] ) ) {

      if( argc != 5 ) {
	 Tcl_SetResult(curFile->interp, insertList[2], TCL_STATIC);
	 return TCL_ERROR;
      }

      if( Tcl_GetInt(curFile->interp, argv[3], &index) != TCL_OK) {
	 Tcl_SetResult(curFile->interp,
		       "Failed to get integer index", TCL_STATIC);
	 return TCL_ERROR;
      }

      if( Tcl_GetInt(curFile->interp, argv[4], &numRows) != TCL_OK) {
	 Tcl_SetResult(curFile->interp,
		       "Failed to get integer numRows", TCL_STATIC);
	 return TCL_ERROR;
      }   
      if( addRowToTable(curFile,index-1,numRows) != TCL_OK ) {
	 return TCL_ERROR;
      } 

   } else if( !strcmp( "image", argv[2] ) ) {

      /*  Write Image Header  */

      int isPrimary;
      if ( argc < 4 || argc > 7 ) {
	 Tcl_SetResult(curFile->interp, insertList[3], TCL_STATIC);
	 return TCL_ERROR;
      }

      /*
       *  Strip off the "-p" flag if present... 
       */
      
      if( !strcmp(argv[3],"-p") ) {
	 isPrimary = 1;
	 for ( i = 3; i < argc - 1; i++ ) {
	    argv[i] = argv[i+1];
	 }
	 argc--;
      } else {
	 isPrimary = 0;
      }
      
      if( fitsPutReqKwds(curFile, isPrimary, IMAGE_HDU, argc, argv) 
	  !=TCL_OK ) {
	 return TCL_ERROR;
      }

   } else if( !strcmp( "table", argv[2] ) ) {

      /*  Write Table Header  */

      int tabType;

      if( argc>3 && !strcmp( "-ascii", argv[3] ) ) {

	 tabType = ASCII_TBL;
	 if( argc < 7 || argc > 11 ) {
	    Tcl_SetResult(curFile->interp, insertList[4], TCL_STATIC);
	    return TCL_ERROR;
	 }
	 for ( i = 3; i < argc - 1; i++ ) {
	    argv[i] = argv[i+1];
	 }
	 argc--;
      
      } else {

	 tabType = BINARY_TBL;
	 if( argc < 6 || argc > 8 ) {
	    Tcl_SetResult(curFile->interp, insertList[4], TCL_STATIC);
	    return TCL_ERROR;
	 }
      
      }

      if( fitsPutReqKwds(curFile, 0, tabType, argc, argv)
	  != TCL_OK ) {
	 return TCL_ERROR;
      }
      
   } else {

      Tcl_SetResult(curFile->interp, "No such insert command", TCL_STATIC);
      return TCL_ERROR;      

   }

   return TCL_OK;
}


/******************************************************************
 *                             Delete
 ******************************************************************/

int fitsTcl_delete( FitsFD *curFile, int argc, char *argv[] )
{
   static char *delKeyList = 
      "delete keyword KeyList\n"
      "       (KeyList can be a mix of keyword names and keyword numbers\n";
   
   static char *delHduList = 
      "delete chdu\n";
   
   static char *delTabList = 
      "delete cols colList\n ";
   
   static char *delRowList = 
      "delete rows -expr expression\n "
      "delete rows firstRow numRows\n ";
   
   int numCols,colNums[FITS_COLMAX],colTypes[FITS_COLMAX],strSize[FITS_COLMAX];
   int fRow, nRows;

   if( argc == 2 ) {
      Tcl_AppendResult(curFile->interp, delKeyList, delHduList, delTabList,
		       delRowList, (char *) NULL);
      return TCL_OK;
   }
   
   if( !strcmp("keyword", argv[2]) ) {

      if( argc != 4 ) {
	 Tcl_SetResult(curFile->interp, delKeyList, TCL_STATIC);
	 return TCL_ERROR;
      }

      if( fitsDeleteKwds(curFile, argv[3] ) != TCL_OK ) {
	 return TCL_ERROR;
      }
      
   } else if( !strcmp("cols", argv[2]) ) {

      if( argc != 4 ) {
	 Tcl_SetResult(curFile->interp, delTabList, TCL_STATIC);
	 return TCL_ERROR;
      }
      if( fitsTransColList( curFile,argv[3],
			    &numCols,colNums,colTypes,strSize) != TCL_OK )
	 return TCL_ERROR;
	 
      if( fitsDeleteCols(curFile, colNums, numCols) != TCL_OK ) {
	 return TCL_ERROR;
      }

   } else if( !strcmp("rows", argv[2]) ) {

      if( argc != 5 ) {
	 Tcl_SetResult(curFile->interp, delRowList, TCL_STATIC);
	 return TCL_ERROR;
      }

      if( !strcmp("-expr", argv[3]) ) {
	 if( fitsDeleteRowsExpr(curFile, argv[4]) != TCL_OK ) {
	    return TCL_ERROR;
	 }
      } else {
	 if( Tcl_GetInt(curFile->interp, argv[3], &fRow) != TCL_OK ) {
	    return TCL_ERROR;
	 }
	 if( Tcl_GetInt(curFile->interp, argv[4], &nRows) != TCL_OK ) {
	    return TCL_ERROR;
	 }
	 if( fitsDeleteRows(curFile, fRow, nRows) != TCL_OK ) {
	    return TCL_ERROR;
	 }
      }
      
   } else if( !strcmp("chdu", argv[2]) ) {

      if( argc != 3 ) {
	 Tcl_SetResult(curFile->interp, delHduList, TCL_STATIC);
	 return TCL_ERROR;
      }
      if( fitsDeleteCHdu(curFile) != TCL_OK ) {
	 return TCL_ERROR;
      }
      
   } else {

      Tcl_SetResult(curFile->interp,
		    "Unrecognized option to delete", TCL_STATIC);
      return TCL_ERROR;

   }
   
   return TCL_OK;
}


/******************************************************************
 *                             Load
 ******************************************************************/

int fitsTcl_load( FitsFD *curFile, int argc, char *argv[] )
{
   static char *loadList = "\n"
      "load column colName ?defaultNull? ?firstElement? - Load a column\n"
      "load vtable colName - Load all elements of a vector column into memory\n"
      "load tblock arrayName colList firstRow numRows colIndex ?felem?\n"
      "             - load a chunk of table and set up an array \"arrayName\"\n"
      "               with indices of (colIndex-1,firstRow-1), etc \n"
      "load image ?slice? ?rotate? - Load a 2D slice of an image into memory\n"
      "                   (rotate: number of 90deg ccw rotations to perform)\n"
      "load irows firstRow lastRow ?slice?  - load mean value of rows\n"
      "load icols firstCol lastCol ?slice?  - load mean value of columns\n"
      "load iblock arrayName firstRow numRows fitsCol numCols ?slice?\n"
      "             - load 2d image slice into an array or memory\n"
      "               if arrayName is --, then a pointer is returned\n"
      "load expr expression ?defaultNull?\n"
      "load keyword -  load the header of CHDU into a hash table \n"
      "load chdu    -  load the CHDU (useful if you move to the CHDU with -s,\n"
      "                which does't load the HDUInfo) \n";

   int numCols,colNums[FITS_COLMAX],colTypes[FITS_COLMAX],strSize[FITS_COLMAX];
   int fRow, nRows;
   int fCol, felem=1;

   if( argc == 2 ) {
      Tcl_SetResult(curFile->interp, loadList, TCL_STATIC);
      return TCL_OK;
   }
   
   if( !strcmp("keyword", argv[2]) ) {
      
      /* Now LOAD the kwds hash table... */

      if( fitsLoadKwds(curFile) != TCL_OK ) {
	 fitsCloseFile((ClientData) curFile);
	 return TCL_ERROR;
      }

   } else if( !strcmp("irows", argv[2]) ) {

      long slice;
      
      if( curFile->hduType != IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not an image", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( argc < 5 ) {
	 Tcl_SetResult(curFile->interp,
		       "FitsHandle load irows firstRow lastRows ?slice?",
		       TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( argc == 5 ) {
	 slice = 1;
      } else {
	 slice = atol(argv[5]);
      }
      
      if( imageRowsMeanToPtr(curFile,
			     atol(argv[3]), /* first row */
			     atol(argv[4]), /* last row*/
			     slice ) != TCL_OK ) {
	 Tcl_AppendResult(curFile->interp,
			  "fitsTcl Error: cannot load irows", NULL);
	 return TCL_ERROR;
      }	
      
   } else if( !strcmp("icols", argv[2]) ) {

      long slice;
      
      if( curFile->hduType != IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not an image", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( argc < 5 ) {
	 Tcl_SetResult(curFile->interp,
		       "FitsHandle load icols firstCol lastCols ?slice?",
		       TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( argc == 5 ) {
	 slice = 1;
      } else {
	 slice = atol(argv[5]);
      }
      
      if( imageColsMeanToPtr(curFile, atol(argv[3]),
			     atol(argv[4]), slice) != TCL_OK ) {
	 Tcl_AppendResult(curFile->interp,
			  "\nfitsTcl Error: cannot load icols",
			  NULL);
	 return TCL_ERROR;
      }	
      
   } else if( !strcmp("iblock", argv[2]) ) {

      long slice=1;
      
      if ( curFile->hduType != IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not an image", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( argc < 8 || argc > 9 ) {
	 Tcl_SetResult(curFile->interp,
		       "FitsHandle load iblock varName firstRow numRows "
		       "firstCol numCols ?slice?", TCL_STATIC); 
	 return TCL_ERROR;
      }

      if( argc > 8 )
	 slice = atol(argv[8]);
      
      if( !strcmp( argv[3], "--" ) )
	 argv[3][0] = '\0';

      if( imageBlockLoad(curFile, argv[3], atol(argv[4]),
			 atol(argv[5]), atol(argv[6]),
			 atol(argv[7]), slice )
	  != TCL_OK ) {
	 return TCL_ERROR;
      }
      
   } else if( !strcmp("tblock", argv[2]) ) {

      if ( curFile->hduType == IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not a table", TCL_STATIC);
	 return TCL_ERROR;
      }

      if (argc < 8 || argc > 10) {
	 Tcl_SetResult(curFile->interp, 
		       "Usage: load tblock arrayName colList "
		       "firstRow numRows firstCol ?felem?", TCL_STATIC);
	 return TCL_ERROR;
      }

      /* parse column list */

      if( fitsTransColList( curFile,argv[4],
			    &numCols,colNums,colTypes,strSize) != TCL_OK )
	 return TCL_ERROR;
      
      /* get the firstRow and numRows */

      if( Tcl_GetInt(curFile->interp, argv[5], &fRow)  != TCL_OK )
	 return TCL_ERROR;
      if( Tcl_GetInt(curFile->interp, argv[6], &nRows) != TCL_OK )
	 return TCL_ERROR;
      if( Tcl_GetInt(curFile->interp, argv[7], &fCol)  != TCL_OK )
	 return TCL_ERROR;

      if( argc>8 ) {  /*  Read felem from very last argument  */
	 if( Tcl_GetInt(curFile->interp, argv[argc-1], &felem) != TCL_OK )
	    return TCL_ERROR;
      }
      
      if( tableBlockLoad(curFile, argv[3], felem, fRow, nRows,
			 fCol, numCols, colNums)   != TCL_OK )
	 return TCL_ERROR;
      
   } else if( !strcmp("image", argv[2]) ) {

      long slice  = 1; 
      int  rotate = 0;
      
      if ( curFile->hduType != IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not an image", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      /* starting element, increment of naxisn[0] x naxisn[1] 
	 to get different frames of a 3d image */

      if( argc == 3 ) {
	 ; /* default to the first frame to allow backward compatible */
      } else if( curFile->CHDUInfo.image.naxes <= 2 ) {
	 ; /* two-d image */
      } else {
	 
	 slice = atol(argv[3]);
	 if( slice < 1 ) {
	    Tcl_SetResult(curFile->interp,
			  "fitsTcl Error: slice starts at 1", TCL_STATIC);
	    return TCL_ERROR;
	 }
	 if( slice > curFile->CHDUInfo.image.naxisn[2] ) {
	    Tcl_SetResult(curFile->interp, 
			  "fitsTcl Error: slice exceeds the 3rd dim",
			  TCL_STATIC);
	    return TCL_ERROR;
	 }

	 if( argc == 5 ) {
	    rotate = atoi(argv[4]);
	    if( rotate<0 || rotate>3 ) {
	       Tcl_SetResult(curFile->interp,
			     "fitsTcl Error: Illegal rotate value",
			     TCL_STATIC);
	       return TCL_ERROR;
	    }
	 }

      }
      
      if( imageGetToPtr(curFile, slice, rotate) != TCL_OK ) {
	 return TCL_ERROR;
      }

   } else if( !strcmp("column", argv[2]) ) {

      char *nullPtr = "NULL";

      if( curFile->hduType == IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not a table", TCL_STATIC);
	 return TCL_ERROR;
      }

      if( argc < 4 || argc > 6 ) {
	 Tcl_SetResult(curFile->interp,
		       "load column colName ?nulValue? ?firstelem?",
		       TCL_STATIC);
	 return TCL_ERROR;
      }

      if( fitsTransColList( curFile, argv[3],
			    &numCols, colNums, colTypes, strSize ) != TCL_OK )
	 return TCL_ERROR;
      if( numCols != 1 ) {
	 Tcl_SetResult(curFile->interp,
		       "Can only load one column at a time", TCL_STATIC);
	 return TCL_ERROR;
      }

      if( argc>4 )
	 nullPtr = argv[4];
      
      if( argc>5 )
	 felem = atol(argv[5]);

      if( tableGetToPtr(curFile, colNums[0], nullPtr, felem) ) {
	 return TCL_ERROR;
      }
      
   } else if( !strcmp("vtable", argv[2]) ) {
      
      if ( curFile->hduType == IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not a table", TCL_STATIC);
	 return TCL_ERROR;
      }

      if( argc < 4 || argc > 5 ) {
	 /* For backwards compatibility, allow for one extra parameter */
	 /* ... formerly the vector size of column                     */
	 Tcl_SetResult(curFile->interp,
		       "load vtable colName", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( fitsTransColList( curFile,argv[3],
			    &numCols,colNums,colTypes,strSize) != TCL_OK )
	 return TCL_ERROR;
      if( numCols != 1 ) {
	 Tcl_SetResult( curFile->interp,
			"Can only load one column at a time", TCL_STATIC );
	 return TCL_ERROR;
      }

      if( vtableGetToPtr(curFile, colNums[0]) ) {
	 return TCL_ERROR;
      }	  
      
   } else if( !strcmp("expr", argv[2]) ) {

      char *nullPtr = "NULL";

      if( curFile->hduType == IMAGE_HDU ) {
	 Tcl_SetResult(curFile->interp,
		       "Current extension is not a table", TCL_STATIC);
	 return TCL_ERROR;
      }

      if( argc < 4 || argc > 5 ) {
	 Tcl_SetResult(curFile->interp, 
		       "Usage: load expr exprStr ?nullVal?", TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( argc > 4 )
	 nullPtr = argv[4];

      if( exprGetToPtr( curFile, argv[3], nullPtr ) ) {
	 return TCL_ERROR;
      }

   } else if( !strcmp("all", argv[2]) || !strcmp("chdu", argv[2]) ) {

      /* load the current hdu */ 

      if( fitsUpdateCHDU(curFile, curFile->hduType) != TCL_OK ) {
	 Tcl_SetResult(curFile->interp,
		       "fitsTcl Error: Cannot update current HDU",
		       TCL_STATIC);
	 return TCL_ERROR;
      }
      
      if( fitsLoadHDU(curFile) != TCL_OK ) {
	 return TCL_ERROR;
      }

   } else {

      Tcl_SetResult(curFile->interp,
		    "Error in fitsTcl: unknown load function", TCL_STATIC);
      return TCL_ERROR;
      
   }

   return TCL_OK;
}


/******************************************************************
 *                             Free
 ******************************************************************/

int fitsTcl_free( FitsFD *curFile, int argc, char *argv[] )
{
   void *databuff;
   char **addList;
   int nAdd;

   if( argc == 2 ) {
      Tcl_SetResult(curFile->interp,
		    "free addressList",
		    TCL_STATIC);
      return TCL_OK;
   }
   
   if( argc>4 ) {
      Tcl_SetResult(curFile->interp, "Too many arguments to free",
		    TCL_STATIC);
      return TCL_ERROR;
   }      

   if( Tcl_SplitList(curFile->interp, argv[argc-1], &nAdd, &addList)
       != TCL_OK ) {
      Tcl_SetResult(curFile->interp,
		    "Cannot parse the address list", TCL_STATIC);
      return TCL_ERROR;
   }

   while( nAdd-- ) {
      databuff = NULL;
      sscanf(addList[nAdd],"%p",&databuff);
      if ( databuff == NULL) {
	 Tcl_SetResult(curFile->interp,
		       "Error interpretting pointer address", TCL_STATIC);
	 ckfree( (char *)addList );
	 return TCL_ERROR;
      }
      ckfree( (char *) databuff);
   }
   ckfree( (char *)addList );

   return TCL_OK;
}


/******************************************************************
 *                             Flush
 ******************************************************************/

int fitsTcl_flush( FitsFD *curFile, int argc, char *argv[] )
{
   int status = 0;

   if ( argc == 2 ) {
      ffflsh(curFile->fptr, 0, &status);
   } else if( argc == 3 ) {
      if( !strcmp(argv[2], "clear") ) {
	 ffflsh(curFile->fptr, 1, &status);
      } else {
	 Tcl_SetResult(curFile->interp, "fitsFile flush ?clear?", TCL_STATIC);
	 return TCL_ERROR;
      }
   } else {
      Tcl_SetResult(curFile->interp, "fitsFile flush ?clear?", TCL_STATIC);
      return TCL_ERROR;
   }
   
   if( status ) {
      Tcl_SetResult(curFile->interp, 
		    "fitsTcl Error: cannot flush file\n", TCL_STATIC);
      dumpFitsErrStack(curFile->interp, status);
      return TCL_ERROR;
   }

   return TCL_OK;
}


/******************************************************************
 *                             Copy
 ******************************************************************/

int fitsTcl_copy( FitsFD *curFile, int argc, char *argv[] )
{
   static char *copyList = "\n"
      "copy filename\n";
   
   if( argc == 2 ) {
      Tcl_SetResult(curFile->interp, copyList, TCL_STATIC);
      return TCL_OK;
   }
   
   if( fitsCopyCHduToFile(curFile, argv[2]) != TCL_OK ) {
      return TCL_ERROR;
   }

   return TCL_OK;
}


/******************************************************************
 *                             Sascii
 ******************************************************************/

int fitsTcl_sascii( FitsFD *curFile, int argc, char *argv[] )
{
   static char *sasciiList =
      "sascii table filename fileMode firstRow numRows colList widthList\n"
      "sascii image filename fileMode firstRow numRows firstCol\n"
      "                                        numCols cellSize ?slice?\n";
   
   int numCols,colNums[FITS_COLMAX],colTypes[FITS_COLMAX],strSize[FITS_COLMAX];
   int fRow, nRows;
   int fCol, nCols, nWdths;
   int cellSize, i;
   char **listWid;

   if( argc == 2 ) {
      Tcl_SetResult(curFile->interp, sasciiList, TCL_STATIC);
      return TCL_OK;
   }      
   
   if( !strcmp("table", argv[2]) ){

      if( argc < 9 || argc > 10 ) {
	 Tcl_SetResult(curFile->interp,
		       "Wrong # of args to 'sascii table'", TCL_STATIC);
	 return TCL_ERROR;
      }  
      
      if( Tcl_GetInt(curFile->interp, argv[5], &fRow) != TCL_OK ) {
	 Tcl_SetResult(curFile->interp, "Cannot get first row", TCL_STATIC);
	 return TCL_ERROR;
      }
      if( Tcl_GetInt(curFile->interp, argv[6], &nRows) != TCL_OK ) {
	 Tcl_SetResult(curFile->interp, "Cannot get number of rows", TCL_STATIC);
	 return TCL_ERROR;
      }

      if( fitsTransColList( curFile,argv[argc-2],
			    &numCols,colNums,colTypes,strSize) != TCL_OK ) {
	 return TCL_ERROR;
      }
      
      if( Tcl_SplitList(curFile->interp, argv[argc-1], &nWdths, &listWid)
	  != TCL_OK ) {
	 Tcl_SetResult(curFile->interp,
		       "Cannot parse the width list", TCL_STATIC);
	 return TCL_ERROR;
      }
      if( nWdths != numCols ) {
	 Tcl_SetResult(curFile->interp,
		       "Cell width array and Column list have different sizes",
		       TCL_STATIC);
	 ckfree( (char*)listWid );
	 return TCL_ERROR;
      }

      for( i=0; i< numCols; i++ ) {
	 if( Tcl_GetInt(curFile->interp, listWid[i], strSize+i) != TCL_OK ) {
	    Tcl_SetResult(curFile->interp,
			  "Unable to parse the width list", TCL_STATIC);
	    ckfree( (char*)listWid );
	    return TCL_ERROR;
	 }
      }
      ckfree( (char*)listWid );

      if( saveTableToAscii( curFile, argv[3], argv[4], 1, fRow, nRows,
			    numCols, colTypes, colNums, strSize ) )
	 return TCL_ERROR;
      
   } else if( !strcmp("image", argv[2]) ) {

      long slice = 1;
      
      if( argc < 10 || argc > 11 ) {
	 Tcl_SetResult(curFile->interp,
		       "Wrong # of args to 'sascii image'", TCL_STATIC);
	 return TCL_ERROR;
      }  
      
      if( argc == 11 ) 
	 slice = atol(argv[10]);
      
      if( Tcl_GetInt(curFile->interp, argv[5], &fRow) != TCL_OK ) {
	 Tcl_SetResult(curFile->interp,
		       "Cannot get first row", TCL_STATIC);
	 return TCL_ERROR;
      }
      if( Tcl_GetInt(curFile->interp, argv[6], &nRows) != TCL_OK ) {
	 Tcl_SetResult(curFile->interp,
		       "Cannot get number of rows", TCL_STATIC);
	 return TCL_ERROR;
      }
      if( Tcl_GetInt(curFile->interp, argv[7], &fCol) != TCL_OK ) {
	 Tcl_SetResult(curFile->interp,
		       "Cannot get first column", TCL_STATIC);
	 return TCL_ERROR;
      }
      if( Tcl_GetInt(curFile->interp, argv[8], &nCols) != TCL_OK ) {
	 Tcl_SetResult(curFile->interp,
		       "Cannot get number of columns", TCL_STATIC);
	 return TCL_ERROR;
      }
      if( Tcl_GetInt(curFile->interp, argv[9], &cellSize) != TCL_OK ) {
	 Tcl_SetResult(curFile->interp,
		       "Cannot get cellSize", TCL_STATIC);
	 return TCL_ERROR;
      }

      if( saveImageToAscii( curFile, argv[3], argv[4], fRow, nRows,
			    fCol, nCols, cellSize, slice ) )
	 return TCL_ERROR;
      
   } else {

      Tcl_SetResult(curFile->interp,
		    "Unknown sascii command", TCL_STATIC);
      return TCL_ERROR;

   }

   return TCL_OK;
}


/******************************************************************
 *                             Sort
 ******************************************************************/

int fitsTcl_sort( FitsFD *curFile, int argc, char *argv[] )
{
   static char *sortList =
      "sort ?-merge? colNameList ?isAscendFlagList? \n";
   
   int numCols,colNums[FITS_COLMAX],colTypes[FITS_COLMAX],strSize[FITS_COLMAX];
   int i;
   int *isAscend;
   char **argPtr;
   char **listPtr;
   int listNum;
   int isMerge = 0;
   
   if( argc == 2 ) { 
      Tcl_SetResult(curFile->interp, sortList, TCL_STATIC);
      return TCL_OK;
   }
   
   if( curFile->hduType == IMAGE_HDU ) {
      Tcl_SetResult(curFile->interp, "Cannot sort an image", TCL_STATIC);
      return TCL_ERROR;
   }      
   
   argc  -= 2;
   argPtr = argv + 2;
   
   if( !strcmp(argPtr[0], "-merge") ) {
      isMerge = 1;
      argc --;
      argPtr ++;
   }
   
   if( fitsTransColList( curFile,argPtr[0],
			 &numCols,colNums,colTypes,strSize) != TCL_OK ) {
      return TCL_ERROR;
   }
   
   isAscend = (int *) ckalloc(numCols*sizeof(int));

   /* if no isAscend specified, set as default ascend */

   if (argc == 1) {

      for (i=0; i < numCols; i++) 
	 isAscend[i] = 1;

   } else {
      
      if( Tcl_SplitList(curFile->interp, argPtr[1],
			&listNum, &listPtr) != TCL_OK ) {
	 ckfree((char *) isAscend);
	 return TCL_ERROR;
      }
      if( listNum != numCols ) {
	 Tcl_SetResult(curFile->interp,
		       "fitsTcl Error: number of flags and columns don't match",
		       TCL_STATIC);
	 ckfree((char *) isAscend);
	 ckfree((char *) listPtr);
	 return TCL_ERROR;
      }
      for (i=0; i< listNum; i++) {
	 if( Tcl_GetInt(curFile->interp, listPtr[i], &isAscend[i]) != TCL_OK ) {
	    ckfree((char*) isAscend);
	    ckfree((char*) listPtr);
	    Tcl_SetResult(curFile->interp,
			  "fitsTcl Error: cannot parse sort flag", TCL_STATIC);
	    return TCL_ERROR;
	 }
      }
      ckfree((char *) listPtr);

   }
   
   if( fitsSortTable(curFile, numCols, colNums,  
		     strSize, isAscend, isMerge) != TCL_OK ) {
      ckfree ((char *) isAscend);
      return TCL_ERROR;
   }

   ckfree ((char *) isAscend);
   return TCL_OK;
}


/******************************************************************
 *                             Add
 ******************************************************************/

int fitsTcl_add( FitsFD *curFile, int argc, char *argv[] )
{
   static char *addColList =
      "add column colName colForm ?expr?\n"
      "    colForm: e.g.\n"
      "    ASCII  Table: A15, I10, E12.5, D20.10, F14.6 ... \n"
      "    BINARY Table: 15A, 1I, 1J, 1E, 1D, 1L, 1X, 1B, 1C, 1M\n";
   static char *addRowList = "add row numRows\n";
   char result[16];
   int numCols,colNums[FITS_COLMAX],colTypes[FITS_COLMAX],strSize[FITS_COLMAX];

   if( argc == 2 ) {
      Tcl_AppendResult(curFile->interp, addColList, addRowList, (char*)NULL);
      return TCL_OK;
   } 

   if( !strcmp(argv[2], "column") ) {
      
      if( argc == 5 ) {

	 if( addColToTable(curFile, FITS_COLMAX, argv[3], argv[4])
	     != TCL_OK ) {
	    return TCL_ERROR;
	 }

      } else if( argc == 6 ) {

	 char *tmpColName;
	 int isNew;
	 
	 strToUpper(argv[3], &tmpColName);
	 if( fitsTransColList(curFile,tmpColName,
			      &numCols,colNums,colTypes,strSize) != TCL_OK ) {

	    /* column name doesn't exist, add a new column*/
	    
	    isNew = 1;
	 } else if( numCols == 1 ) {
	    isNew = 0;
	 } else {
	    Tcl_SetResult(curFile->interp,
			  "Can only add one column at a time", TCL_STATIC);
	    ckfree((char *) tmpColName);
	    return TCL_ERROR;
	 }
	 ckfree((char *) tmpColName);
	 
	 if( fitsCalculateColumn(curFile, argv[3],
				 ( strcmp(argv[4],"default") ? argv[4] : NULL ),
				 argv[5]) != TCL_OK ) {
	    return TCL_ERROR;
	 }
	 sprintf(result,"%d",isNew);
	 Tcl_SetResult(curFile->interp, result, TCL_VOLATILE);

      } else {

	 Tcl_SetResult(curFile->interp, addColList, TCL_STATIC);
	 return TCL_ERROR;

      }

   } else if( !strcmp(argv[2], "row") ) {

      int numRows;

      if( argc != 4 ) {
	 Tcl_SetResult(curFile->interp, addRowList, TCL_STATIC);
	 return TCL_ERROR;
      }
      if( Tcl_GetInt(curFile->interp, argv[3], &numRows) != TCL_OK) {
	 Tcl_SetResult(curFile->interp,
		       "Failed to get numRows parameter", TCL_STATIC);
	 return TCL_ERROR;
      }       
      if( addRowToTable(curFile, curFile->CHDUInfo.table.numRows,  
			numRows) != TCL_OK ) {
	 return TCL_ERROR;
      }

   } else {

      Tcl_SetResult(curFile->interp, "Unknown add command", TCL_STATIC);
      return TCL_ERROR;

   }   

   return TCL_OK;
}


/******************************************************************
 *                             Append
 ******************************************************************/

int fitsTcl_append( FitsFD *curFile, int argc, char *argv[] )
{
   static char *appendList = "\n"
      "append filename \n"
      "       -- append the chdu to another file\n";

   if( argc < 3 ) {
      Tcl_SetResult(curFile->interp, appendList, TCL_STATIC);
      return TCL_OK;
   }

   if( fitsAppendCHduToFile(curFile, argv[2]) != TCL_OK ) {
      return TCL_ERROR;
   } 

   return TCL_OK;
}


/******************************************************************
 *                             Create
 ******************************************************************/

int fitsTcl_create( FitsFD *curFile, int argc, char *argv[] )
{
   static char *createList= "\n"
      "create 2dhisto filename {colList} {xmin xmax xbin} {ymin ymax ybin} ?rows?\n"
      "       1dhisto filename {colList} {xmin xmax xbin} ?row?\n";
   
   int numRange, range[FITS_MAXRANGE][2];
   int fRow, lRow;
   char errMsg[256];

   if( argc == 2 ) {
      Tcl_SetResult(curFile->interp, createList, TCL_STATIC);
      return TCL_OK;
   }
   fRow     = 1;
   lRow     = curFile->CHDUInfo.table.numRows;
   numRange = 1;    
   
   if( !strcmp("2dhisto", argv[2]) ) {
      
      if( argc < 7) {
	 Tcl_SetResult(curFile->interp, "Wrong # of args to 'create 2dhisto'",
		       TCL_STATIC);
	 return TCL_ERROR;
      } else if ( argc > 7) {
	 if( fitsParseRange(argv[7],&numRange,range,FITS_MAXRANGE,
			    1, curFile->CHDUInfo.table.numRows,errMsg)
	     != TCL_OK ) {
	    Tcl_SetResult(curFile->interp,
			  "Error parsing row range:\n", TCL_STATIC);
	    Tcl_AppendResult(curFile->interp, errMsg, (char*)NULL);
	    return TCL_ERROR;
	 }
	 if( numRange != 1 ) {
	    Tcl_SetResult(curFile->interp,
			  "'create' only supports a single row range",
			  TCL_STATIC);
	    return TCL_ERROR;
	 }
	 fRow = range[0][0];
	 lRow = range[0][1];
      } else {
	 ;
      }
      
      if( fitsWrite2DHisto(curFile, 
			   argv[3], /* file name */
			   argv[4], /* XY?Z? column list */
			   argv[5], /* {xmin xmax xbinsize} */
			   argv[6], /* {ymin ymax ybinsize} */
			   fRow,    /* first row */
			   lRow)    /* last row */
	  != TCL_OK ) {
	 return TCL_ERROR;
      }
      
   } else if( !strcmp("1dhisto", argv[2]) ) {
      
      if( argc < 6 ) {
	 Tcl_SetResult(curFile->interp, "Wrong # of args to 'create 2dhisto'",
		       TCL_STATIC);
	 return TCL_ERROR;
      } else if( argc > 6 ) {
	 if( fitsParseRange(argv[6],&numRange,range,FITS_MAXRANGE,
			    1, curFile->CHDUInfo.table.numRows,errMsg)
	     != TCL_OK ) {
	    Tcl_SetResult(curFile->interp,
			  "Error parsing row range:\n", TCL_STATIC);
	    Tcl_AppendResult(curFile->interp, errMsg, (char*)NULL);
	    return TCL_ERROR;
	 }
	 if( numRange != 1 ) {
	    Tcl_SetResult(curFile->interp,
			  "'create' only supports a single row range",
			  TCL_STATIC);
	    return TCL_ERROR;
	 }
	 fRow = range[0][0];
	 lRow = range[0][1];
      } else {
	 ;
      }
      
      if( fitsWrite1DHisto(curFile, 
			   argv[3], /* file name */
			   argv[4], /* column  */
			   argv[5], /* {xmin xmax xbinsize} */
			   fRow,    /* first row */
			   lRow)    /* last row */
	  != TCL_OK ) {
	 return TCL_ERROR;
      }
      
   } else {
      Tcl_SetResult(curFile->interp, "Unknown 'create' command", TCL_STATIC);
      return TCL_ERROR;
   }

   return TCL_OK;
}


/******************************************************************
 *                             Checksum
 ******************************************************************/

int fitsTcl_checksum( FitsFD *curFile, int argc, char *argv[] )
{
   static char *checksumList="\n"
      "checksum verify\n"
      "checksum update\n";
   
   char result[16];
   int datastatus = 0;
   int hdustatus  = 0;
   int status     = 0;

   if( argc < 3 ) {
      Tcl_SetResult(curFile->interp, checksumList, TCL_STATIC);
      return TCL_OK;
   } 
   
   if( !strcmp("verify", argv[2]) ) {
      
      /* verify the checksum keyword. */
      /* return 1 OK, 0 checksum keyword not present, -1 wrong */

      if( ffvcks(curFile->fptr, &datastatus, &hdustatus, &status) ) {
	 dumpFitsErrStack(curFile->interp, status);
	 return TCL_ERROR;
      }
      /*  Return "minimum" checksum status  */
      sprintf(result, "%d", (hdustatus<datastatus ? hdustatus : datastatus) );
      Tcl_SetResult(curFile->interp, result, TCL_VOLATILE);
      
   } else if( !strcmp("update", argv[2]) ) {

      if( ffpcks(curFile->fptr, &status) ) {
	 dumpFitsErrStack(curFile->interp, status);
	 return TCL_ERROR;
      }

      if( fitsUpdateFile(curFile) != TCL_OK ) {
	 return TCL_ERROR;
      }

   } else {

      Tcl_SetResult(curFile->interp, "Unknown checksum option", TCL_STATIC);
      return TCL_ERROR;

   }
   
   return TCL_OK;
}
